from typing import Optional

import click

from mlfoundry.login import login as login_


@click.command(short_help="Store API key in the local file system")
@click.option(
    "--tracking_uri",
    type=str,
    default=None,
    help="Tracking server host",
)
@click.option(
    "--relogin",
    is_flag=True,
    show_default=True,
    default=False,
    help="Overwrite existing API key for the given `tracking_uri`",
)
def login(tracking_uri: Optional[str], relogin: bool):
    login_(tracking_uri=tracking_uri, relogin=relogin)



import enum
import typing

from mlfoundry.log_types.pydantic_base import PydanticBase


class StatsSource(enum.Enum):
    WHYLOGS = "WHYLOGS"


class DatasetStatsRunLog(PydanticBase):
    value: typing.Dict
    stats_source: StatsSource

    @staticmethod
    def get_log_type() -> str:
        return "dataset/stats"



from sqlalchemy import JSON, BigInteger, Column, DateTime, Index, Integer, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class InferenceLog(Base):
    __tablename__ = "mlfoundry_inference_log"
    # https://docs.sqlalchemy.org/en/13/dialects/sqlite.html#allowing-autoincrement-behavior-sqlalchemy-types-other-than-integer-integer
    id = Column(
        BigInteger().with_variant(Integer, "sqlite"),
        primary_key=True,
        autoincrement=True,
    )
    # Identifying the model
    model_name = Column(String(256), nullable=False)
    model_version = Column(String(64), nullable=False)
    # identifying the inference payload
    inference_id = Column(String(36), nullable=False)
    features = Column(JSON, nullable=False)
    predictions = Column(JSON, nullable=False)
    occurred_at = Column(DateTime(), nullable=False)

    raw_data = Column(JSON)
    actuals = Column(JSON)

    __table_args__ = (
        Index(
            "mlfoundry_inference_log_model_name_inference_id",
            "model_name",
            "inference_id",
        ),
    )



# The logic is copied from
# https://github.com/truefoundry/mlfoundry-server/blob/c2334ace8f35322cd5e50c275b8df08327688c01/mlflow/tracking/client.py#L1188
import logging
from typing import Union

import numpy as np

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.image.constants import MISSING_PILLOW_PACKAGE_MESSAGE
from mlfoundry.run_utils import get_module

logger = logging.getLogger(__name__)


def _normalize_to_uint8(x):
    # Based on: https://github.com/matplotlib/matplotlib/blob/06567e021f21be046b6d6dcf00380c1cb9adaf3c/lib/matplotlib/image.py#L684

    is_int = np.issubdtype(x.dtype, np.integer)
    low = 0
    high = 255 if is_int else 1
    if x.min() < low or x.max() > high:
        msg = (
            "Out-of-range values are detected. "
            "Clipping array (dtype: '{}') to [{}, {}]".format(x.dtype, low, high)
        )
        logger.warning(msg)
        x = np.clip(x, low, high)

    # float or bool
    if not is_int:
        x = x * 255

    return x.astype(np.uint8)


def _convert_numpy_to_pil_image(image: np.ndarray) -> "PIL.Image.Image":
    pil_image_module = get_module(
        module_name="PIL.Image",
        required=True,
        error_message=MISSING_PILLOW_PACKAGE_MESSAGE,
    )
    valid_data_types = {
        "b": "bool",
        "i": "signed integer",
        "u": "unsigned integer",
        "f": "floating",
    }

    if image.dtype.kind not in valid_data_types.keys():
        raise TypeError(
            "Invalid array data type: '{}'. Must be one of {}".format(
                image.dtype, list(valid_data_types.values())
            )
        )

    if image.ndim not in [2, 3]:
        raise ValueError(
            "`image` must be a 2D or 3D array but got a {}D array".format(image.ndim)
        )

    if (image.ndim == 3) and (image.shape[2] not in [1, 3, 4]):
        raise ValueError(
            "Invalid channel length: {}. Must be one of [1, 3, 4]".format(
                image.shape[2]
            )
        )

    # squeeze a 3D grayscale image since `Image.fromarray` doesn't accept it.
    if image.ndim == 3 and image.shape[2] == 1:
        image = image[:, :, 0]

    image = _normalize_to_uint8(image)

    return pil_image_module.fromarray(image)


def normalize_image(
    image: Union["PIL.Image.Image", "numpy.ndarray"]
) -> "PIL.Image.Image":
    pil_image_module = get_module(
        module_name="PIL.Image",
        required=True,
        error_message=MISSING_PILLOW_PACKAGE_MESSAGE,
    )
    if isinstance(image, pil_image_module.Image):
        return image
    if isinstance(image, np.ndarray):
        return _convert_numpy_to_pil_image(image)
    raise MlFoundryException(
        f"image should be of type PIL Image/np.ndarray got type {type(image)}"
    )



from getpass import getpass
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urljoin, urlparse

from tinynetrc import Netrc

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.run_utils import resolve_tracking_uri

CREDENTIALS_DIR = Path.home() / ".mlfoundry"
CREDENTIALS_FILE = CREDENTIALS_DIR / "credentials.netrc"


def login(
    tracking_uri: Optional[str] = None,
    relogin: bool = False,
    api_key: Optional[str] = None,
) -> bool:
    """Save API key in local file system for a given `tracking_uri`.

    Args:
        tracking_uri (Optional[str], optional): tracking_uri for the given API key
        relogin (bool, optional): Overwrites the existing API key for the `tracking_uri` if
            set to `True`. If set to `False` and an API key is already present for
            the given `tracking_uri`, then the existing API key is kept untouched.
            Default is `False`.
        api_key (Optional[str], optional): The API key for the given `tracking_uri`.
            If `api_key` is not passed, this function prompts for the API key.

    Returns:
        bool: Returns `True` if the given API Key was persisted.
    """
    tracking_uri = resolve_tracking_uri(tracking_uri)
    return _Login(tracking_uri).login(relogin=relogin, api_key=api_key)


def get_stored_api_key(tracking_uri: str) -> Optional[str]:
    """Get API key for a `tracking_uri` from the local file system.

    Args:
        tracking_uri (str): tracking_uri

    Returns:
        Optional[str]: The stored API key for the `tracking_uri`. If not present
            then this function returns `None`.
    """
    return _Login(tracking_uri).fetch_api_key()


def _prompt_api_key(tracking_uri: str) -> str:
    api_key_link = urljoin(tracking_uri, "/settings")
    print(f"Please get your API key from {api_key_link}")
    api_key = getpass("Paste your API key and hit enter:")
    return api_key


class _Login:
    def __init__(
        self,
        tracking_uri: str,
        credentials_file_path: Union[Path, str] = CREDENTIALS_FILE,
    ):
        self.tracking_uri = tracking_uri
        self.tracking_host = urlparse(tracking_uri).netloc
        if not self.tracking_host:
            raise MlFoundryException(f"invalid url: {tracking_uri}")

        self.credentials_file_path = Path(credentials_file_path).absolute()

        if not self.credentials_file_path.exists():
            self.credentials_file_path.parent.mkdir(exist_ok=True, parents=True)
            self.credentials_file_path.touch(exist_ok=True)

    def fetch_api_key(self) -> Optional[str]:
        return Netrc(self.credentials_file_path)[self.tracking_host]["password"]

    def login(self, relogin: bool = False, api_key: Optional[str] = None) -> bool:
        if relogin:
            api_key = api_key or _prompt_api_key(self.tracking_uri)
            self._save_api_key(api_key)
            return True
        existing_api_key = self.fetch_api_key()
        if not existing_api_key:
            api_key = api_key or _prompt_api_key(self.tracking_uri)
            self._save_api_key(api_key)
            return True
        print(
            "API key is already configured.\n"
            "Please use `mlfoundry login --relogin` or `mlfoundry.login(relogin=True)`"
            "to force relogin"
        )
        return False

    def _save_api_key(self, api_key: str):
        print(f"Writing API key at {self.credentials_file_path}")
        with Netrc(self.credentials_file_path) as netrc:
            netrc[self.tracking_host]["password"] = api_key



import posixpath
import re

IMAGE_LOG_DIR = posixpath.join("mlf", "run_logs", "images")
MEDIA_DIR = posixpath.join(IMAGE_LOG_DIR, "media")
DEFAULT_IMAGE_FORMAT = "png"
MISSING_PILLOW_PACKAGE_MESSAGE = (
    "We need PIL package to save image.\nTo install, run `pip install pillow`"
)
IMAGE_KEY_REGEX = re.compile(r"^[a-zA-Z0-9-_]+$")



import os
import posixpath
import shutil
import tempfile
from pathlib import Path
from typing import Dict, List, Optional, Union
from uuid import uuid4

from mlflow.entities import RunLog

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.log_types.image.constants import (
    DEFAULT_IMAGE_FORMAT,
    IMAGE_KEY_REGEX,
    IMAGE_LOG_DIR,
    MEDIA_DIR,
    MISSING_PILLOW_PACKAGE_MESSAGE,
)
from mlfoundry.log_types.image.image_normalizer import normalize_image
from mlfoundry.log_types.image.types import BoundingBoxGroups, ClassGroups
from mlfoundry.log_types.pydantic_base import PydanticBase
from mlfoundry.run_utils import get_module

DataOrPathType = Union[str, Path, "numpy.ndarray", "PIL.Image.Image"]
ClassGroupsType = Dict[str, Union[str, List[str]]]
BBoxGrouptype = Dict[str, List[Dict]]


def validate_key_name(key: str):
    if not key or not IMAGE_KEY_REGEX.match(key):
        raise MlFoundryException(
            f"Invalid run image key: {key} should only contain alphanumeric, hyphen or underscore"
        )


def get_run_log_file_name(key: str, step: int):
    return f"{key}-{step}-{uuid4()}.json"


class Image:
    def __init__(
        self,
        data_or_path: DataOrPathType,
        caption: Optional[str] = None,
        class_groups: Optional[ClassGroupsType] = None,
        bbox_groups: Optional[BBoxGrouptype] = None,
    ):
        """Represent and log image using this class in `mlfoundry`.

        You can initialize `mlfoundry.Image` by either by using a local path
        or you can use a numpy array / PIL.Image object.

        If you are using numpy array, we only support the following data types,
        - bool
        - integer [0 - 255]
        - unsigned integer [0 - 255]
        - float [0.0 - 1.0]

        Any out of range value will be clipped.

        As for array shape/dim, we follow the following structures,
        - H x W (Grayscale)
        - H x W x 1 (Grayscale)
        - H x W x 3 (an RGB channel order is assumed)
        - H x W x 4 (an RGBA channel order is assumed)

        `PIL` package is required to log images. To install the `PIL` package,
        run `pip install pillow`.

        We can also log class names and bounding boxes associated with the image.
        Class names and bounding boxes should be always grouped under `actuals` or
        `predictions`. For example, if we have an image where the ground truth class
        is "cat" and predicted class is "dog", we can represent it like,

        ```python
        mlfoundry.Image(
            data_or_path=imarray,
            class_groups={"actuals": "dog", "predictions": "cat"}
        )
        ```

        You can define a bounding box using the following dictionary structure,
        ```python
        {
            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30}, # required, defines the position of the bounding box
                                                                             # (min_x, min_y) defines the top left and
                                                                             # (max_x, max_y) defines the bottom right corner of the box.
            "class_name": "dog", # required, the class name of the bounding box
            "caption": "dog", # optional, the caption of the bounding box.
                              # If not passed, the class name is set as caption.
        }
        ```

        Args:
            data_or_path (Union[str, Path, "numpy.ndarray", "PIL.Image.Image"]):
                Either the local path or the image object (Numpy array or PIL Image).
            caption (Optional[str], optional): A string caption or label for the image.
            class_groups (Optional[Dict[str, Union[str, List[str]]]], optional):
                Class names associated with the image. Expects class name(s) grouped by
                `predictions` or `actuals`.
            bbox_groups (Optional[Dict[str, List[Dict]]], optional): Bounding boxes
                associated with the image. Expects bounding boxes grouped by `predictions`
                or `actuals`.

        Examples:
        ### Logging images with caption and class names

        ```python
        import mlfoundry
        import numpy as np

        client = mlfoundry.get_client()
        run = client.create_run(
            project_name="my-classification-project",
        )

        imarray = np.random.randint(low=0, high=256, size=(100, 100, 3))

        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": "dog", "predictions": "cat"},
            ),
        }
        run.log_images(images_to_log, step=1)

        run.end()
        ```

        ### Logging images for a multi-label classification problem

        ```python
        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                class_groups={"actuals": ["dog", "human"], "predictions": ["cat", "human"]},
            ),
        }

        run.log_images(images_to_log, step=1)
        ```

        ### Logging images with bounding boxes

        ```python
        images_to_log = {
            "logged-image-array": mlfoundry.Image(
                data_or_path=imarray,
                caption="testing image logging",
                bbox_groups={
                    "predictions": [
                        {
                            "position": {"min_x": 5, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "cat",
                        }
                    ],
                    "actuals": [
                        {
                            "position": {"min_x": 15, "min_y": 5, "max_x": 20, "max_y": 30},
                            "class_name": "dog",
                            "caption": "dog",
                        }
                    ],
                },
            ),
        }

        run.log_images(images_to_log, step=1)
        ```
        """
        self._run = None

        self._caption = None
        self._class_groups = None
        self._bbox_groups = None

        self._image = None
        self._image_artifact_path = None
        self._local_image_path = None

        self._init_image(data_or_path)
        self._init_metadata(
            caption=caption, class_groups=class_groups, bbox_groups=bbox_groups
        )

    @property
    def image(self) -> "PIL.Image.Image":
        if self._image is None:
            raise MlFoundryException("Image is not initialized")
        return self._image

    @property
    def caption(self) -> Optional[str]:
        return self._caption

    @property
    def class_groups(self) -> Optional[ClassGroups]:
        return self._class_groups

    @property
    def bbox_groups(self) -> Optional[BoundingBoxGroups]:
        return self._bbox_groups

    @property
    def run(self) -> "mlfoundry.MlFoundryRun":
        if self._run is None:
            raise MlFoundryException("Image is not bound to a run yet.")
        return self._run

    @property
    def image_artifact_path(self) -> str:
        if self._image_artifact_path is not None:
            return self._image_artifact_path

        if self._local_image_path is None:
            image_artifact_path = self._serialize_and_save_image_as_artifact()
        else:
            image_artifact_path = self._copy_local_image_as_artifact()

        self._image_artifact_path = image_artifact_path
        return self._image_artifact_path

    def _serialize_and_save_image_as_artifact(
        self, artifact_path: str = MEDIA_DIR
    ) -> str:
        file_name = f"{uuid4()}.{DEFAULT_IMAGE_FORMAT}"
        image_artifact_path = posixpath.join(artifact_path, file_name)
        with tempfile.TemporaryDirectory() as local_dir:
            local_path = os.path.join(local_dir, file_name)
            self.image.save(local_path)
            self.run.mlflow_client.log_artifact(
                run_id=self.run.run_id,
                local_path=local_path,
                artifact_path=artifact_path,
            )
        return image_artifact_path

    def _copy_local_image_as_artifact(self, artifact_path: str = MEDIA_DIR) -> str:
        file_name = os.path.basename(self._local_image_path)
        new_file_name = f"{uuid4()}-{file_name}"
        image_artifact_path = posixpath.join(artifact_path, new_file_name)
        with tempfile.TemporaryDirectory() as local_dir:
            new_local_image_path = os.path.join(local_dir, new_file_name)
            shutil.copy2(self._local_image_path, new_local_image_path)
            self.run.mlflow_client.log_artifact(
                run_id=self.run.run_id,
                local_path=new_local_image_path,
                artifact_path=artifact_path,
            )
        return image_artifact_path

    def _bind_to_run(self, run: "mlfoundry.MlFoundryRun"):
        if self._run is None:
            self._run = run
            return
        if self._run.run_id != run.run_id:
            raise MlFoundryException(
                f"Image is already bound to run {self._run.run_id}"
            )

    def _init_image(self, data_or_path: DataOrPathType):
        pil_image_module = get_module(
            module_name="PIL.Image",
            required=True,
            error_message=MISSING_PILLOW_PACKAGE_MESSAGE,
        )
        if isinstance(data_or_path, (str, Path)):
            self._local_image_path = os.path.abspath(data_or_path)
            with pil_image_module.open(data_or_path) as image:
                image.load()
                self._image = image
        else:
            self._image = normalize_image(data_or_path)

    def _init_metadata(
        self,
        caption: Optional[str],
        class_groups: Optional[ClassGroupsType],
        bbox_groups: Optional[Dict],
    ):
        if caption is not None:
            self._caption = str(caption)

        if class_groups:
            self._class_groups = ClassGroups.parse_obj(class_groups)

        if bbox_groups:
            self._bbox_groups = BoundingBoxGroups.parse_obj(bbox_groups)

    def _to_dict(self) -> Dict:
        dict_ = {}
        dict_["caption"] = self._caption
        dict_["image_artifact_path"] = self.image_artifact_path
        dict_["class_groups"] = (
            self.class_groups.to_dict() if self.class_groups is not None else None
        )
        dict_["bbox_groups"] = (
            self.bbox_groups.to_dict() if self.bbox_groups is not None else None
        )

        return dict_

    def to_run_log(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
        timestamp: Optional[int] = None,
    ) -> RunLog:
        validate_key_name(key)

        self._bind_to_run(run)

        return ImageRunLogType(value=self._to_dict()).to_run_log_as_artifact(
            key=key,
            run_id=self.run.run_id,
            mlflow_client=self.run.mlflow_client,
            file_name=get_run_log_file_name(key=key, step=step),
            artifact_path=IMAGE_LOG_DIR,
            step=step,
            timestamp=timestamp,
        )

    def save(
        self,
        run: "mlfoundry.MlFoundryRun",
        key: str,
        step: int = 0,
        timestamp: Optional[int] = None,
    ):
        run_log = self.to_run_log(
            run=run,
            key=key,
            step=step,
            timestamp=timestamp,
        )
        self.run.mlflow_client.insert_run_logs(
            run_uuid=self.run.run_id, run_logs=[run_log]
        )


class ImageRunLogType(PydanticBase):
    value: Dict

    @staticmethod
    def get_log_type() -> str:
        return "image"



import logging
import queue
import threading
import time
from typing import Any, Dict

from mlflow.entities import Metric

from mlfoundry.background import utils as bg_utils
from mlfoundry.background.events import Event, LogMetricsEvent

logger = logging.getLogger(__name__)


class Interface(object):
    """
    Interface that facilitates communication between producing threads and consuming threads via an event queue
    """

    def __init__(
        self,
        run_id: str,
        event_queue: "queue.Queue[Event]",
    ):
        """
        Args:
            event_queue: queue to hold events
        """
        self.run_id = run_id
        self.event_queue = event_queue
        self._event_queue_closed = False

    def log_metrics(self, metric_dict: Dict[str, Any], step: int = 0) -> None:
        """
        Publish `metric_dict` as event to event queue which would be logged asynchronously

        Args:
            metric_dict: dict mapping key (string) to values (mostly float, int) to log as metrics
            step: step number or iteration number, defaults to zero
        """
        # TODO (chiragjn): Yuck! duplication between Run instance and this interface
        #                  We should start moving all logic to interface and make public log methods
        #                  on Run instance act as proxies with the option to be sync / async.
        if not metric_dict:
            return
        timestamp = int(time.time() * 1000)
        metrics_arr = [
            Metric(key, value, timestamp, step=step)
            for key, value in metric_dict.items()
        ]
        event = LogMetricsEvent(
            producer=threading.currentThread().getName(), metrics=metrics_arr
        )
        self.put_event(event)

    def put_event(self, event: Event) -> None:
        """
        Put the given `event` to event queue if the queue is not yet closed.

        Args:
            event: instance of `Event` or subclass of `Event`
        """
        if self._event_queue_closed:
            logger.warning(
                bg_utils.make_log(
                    f"put_event called with {event!r} after close. Dropping event"
                )
            )
            return
        logger.debug(bg_utils.make_log(f"enqueue_event {event!r}"))
        self.event_queue.put(event)

    def get_event(self, **get_kwargs) -> Event:
        """
        Pop and get `Event` from event queue

        Args:
            **get_kwargs: kwargs to forward to `queue.Queue.get`

        Returns:
            Event: instance of `Event` or subclass of `Event`

        Raises:
            queue.Empty: if the event queue is empty
        """
        try:
            event = self.event_queue.get(**get_kwargs)
            logger.debug(bg_utils.make_log(f"dequeue_event {event!r}"))
            return event
        except queue.Empty:
            raise

    def close(self) -> None:
        """
        Mark the event queue as closed to stop accepting more events
        """
        self._event_queue_closed = True



from mlfoundry.enums import FileFormat
from mlfoundry.log_types.pydantic_base import PydanticBase


class DatasetArtifact(PydanticBase):
    artifact_path: str

    @staticmethod
    def get_log_type() -> str:
        return "dataset"


class DatasetArtifactFeatures(PydanticBase):
    artifact_path: str
    format: FileFormat

    @staticmethod
    def get_log_type() -> str:
        return "dataset/features"


class DatasetArtifactActuals(PydanticBase):
    artifact_path: str
    format: FileFormat

    @staticmethod
    def get_log_type() -> str:
        return "dataset/actuals"


class DatasetArtifactPredictions(PydanticBase):
    artifact_path: str
    format: FileFormat

    @staticmethod
    def get_log_type() -> str:
        return "dataset/predictions"


class DatasetArtifactFeaturesProfile(PydanticBase):
    artifact_path: str

    @staticmethod
    def get_log_type() -> str:
        return "dataset/features/profile"


class DatasetArtifactActualsProfile(PydanticBase):
    artifact_path: str

    @staticmethod
    def get_log_type() -> str:
        return "dataset/actuals/profile"


class DatasetArtifactPredictionsProfile(PydanticBase):
    artifact_path: str

    @staticmethod
    def get_log_type() -> str:
        return "dataset/predictions/profile"



import tempfile
import typing

import mlflow
from packaging.version import parse

from mlfoundry.exceptions import MlFoundryException
from mlfoundry.frameworks.base_registry import BaseRegistry


def get_tf():
    try:
        import tensorflow as tf

        return tf
    except ImportError as ex:
        raise ImportError(
            "Tensorflow package is required to use TfSavedModelArtifact."
        ) from ex


# right now only tensorflow 2 is supported
# I will take up if tf1 if there is any requirement
def assert_tf_2():
    tf = get_tf()
    version_string = tf.__version__
    major_version = parse(version_string).release[0]
    if major_version < 2:
        raise MlFoundryException(
            f"Only Tensorflow 2 is supported. Found tensorflow version {version_string}"
        )


class TensorflowRegistry(BaseRegistry):
    def log_model(
        self,
        model,
        artifact_path: str,
        signatures=None,
        options=None,
        tf_signature_def_key: typing.Optional[str] = None,
    ):
        assert_tf_2()
        tf = get_tf()
        tf_signature_def_key = (
            tf.saved_model.DEFAULT_SERVING_SIGNATURE_DEF_KEY
            if tf_signature_def_key is None
            else tf_signature_def_key
        )
        with tempfile.TemporaryDirectory() as local_dir:
            tf.saved_model.save(
                obj=model,
                export_dir=local_dir,
                signatures=signatures,
                options=options,
            )
            mlflow.tensorflow.log_model(
                tf_saved_model_dir=local_dir,
                tf_meta_graph_tags=None,
                tf_signature_def_key=tf_signature_def_key,
                artifact_path=artifact_path,
            )

    def load_model(self, model_file_path: str, **kwargs):
        assert_tf_2()
        return mlflow.tensorflow.load_model(model_file_path)



import logging
import os
import posixpath
import re
import tempfile
import typing
import uuid
from distutils.dir_util import copy_tree

import mlflow
from mlflow.entities import RunLog, RunStatus
from mlflow.tracking import MlflowClient

from mlfoundry.enums import ModelFramework
from mlfoundry.exceptions import MlFoundryException
from mlfoundry.frameworks import get_model_registry
from mlfoundry.log_types import ModelArtifactRunLog

MODEL_PREFIX = posixpath.join("mlf", "run-logs", "models")
MODEL_NAME_REGEX = re.compile(r"^[a-zA-Z0-9-_]*$")
logger = logging.getLogger(__name__)

# Right now we do not have concept of a model name.
# Now while saving the model as a run log, we need to give a key name.
# Right now I am hardcoding it to "model".
# This is not exposed to user at all.
# We should not change this default value at all.
DEFAULT_MODEL_NAME = "model"


def load_model_from_local_path(local_path: str):
    # this will fail if the model was not serialized
    # using tfy-mlflow-client
    # test all flavors from here

    model = mlflow.models.Model.load(local_path)
    for flavor in model.flavors.keys():
        try:
            framework = ModelFramework(flavor)
        except Exception:
            continue
        registry = get_model_registry(framework)
        return registry.load_model(local_path)
    raise MlFoundryException(f"unsupported frameworks {model.flavors.keys()}")


class ModelDriver:
    """ModelDriver."""

    def __init__(self, mlflow_client: MlflowClient, run_id: str):
        """__init__.

        :param mlflow_client:
        :type mlflow_client: MlflowClient
        :param run_id:
        :type run_id: str
        """
        self.mlflow_client: MlflowClient = mlflow_client
        self.run_id: str = run_id

    def _get_latest_run_log(
        self,
        model_name: str,
    ) -> typing.Optional[RunLog]:
        """
        Get latest run log under a run for a model (one with the largest step)
        If there are no run log then this function will return None.

        :param model_name: Name of the model. This will be just DEFAULT_MODEL_NAME
                           for now.
        :type model_name: str
        :rtype: typing.Optional[RunLog]
        """
        run_log = self.mlflow_client.get_latest_run_log(
            run_uuid=self.run_id,
            key=model_name,
            log_type=ModelArtifactRunLog.get_log_type(),
        )
        return run_log

    def _get_run_log(self, model_name: str, step: int) -> typing.Optional[RunLog]:
        """
        Get run log under a run for a model
        If there are no run log then this function will return None.

        :param model_name: Name of the model. This will be just DEFAULT_MODEL_NAME
                           for now.
        :type model_name: str
        :param step: step number (for which run_log is to be fetched)
        :rtype: typing.Optional[RunLog]
        """
        run_logs = self.mlflow_client.list_run_logs(
            run_uuid=self.run_id,
            key=model_name,
            log_type=ModelArtifactRunLog.get_log_type(),
            steps=[step],
        )

        if len(run_logs) == 0:
            return None
        return run_logs[0]

    def log_model(self, model, framework: str, step: int = 0, **kwargs):
        """
        Serialize and log a model for the current `run`. Each logged model is
            associated with a step. After logging model at a particular step
            we cannot overwrite it.

        Args:
            model: The model object
            framework (Union[enums.ModelFramework, str]): Model Framework. Ex:- pytorch,
                sklearn, tensorflow etc. The full list of supported frameworks can be
                found in `mlfoundry.enums.ModelFramework`.
            step (int, optional): step/iteration at which the model is being logged
                If not passed, `0` is set as the `step`.
            kwargs: Keyword arguments to be passed to the serializer.
        """
        model_name = DEFAULT_MODEL_NAME
        run_log = self._get_run_log(model_name, step)
        if run_log is not None:
            raise MlFoundryException(
                f"Model already logged with step: {step}, cannot overwrite."
            )
        framework = ModelFramework(framework)
        artifact_path = posixpath.join(
            MODEL_PREFIX, f"{model_name}-{step}-{uuid.uuid4().hex}"
        )
        # TODO (chiragjn): This is hack to temporarily use the mlflow.fluent API because
        #                  .log_model relies on run_id being stored in mlflow's global run stack
        created_run = None
        try:
            # Add to mlflow.fluent's global stack, hoping no other fluent run is running parallely
            created_run = mlflow.start_run(
                run_id=self.run_id,
                nested=True,  # necessary to avoid messing with existing run stack
            )
            get_model_registry(framework).log_model(
                model, artifact_path=artifact_path, **kwargs
            )
        finally:
            if created_run:
                # Remove from mlflow.fluent's global stack but keeping the run in Running mode
                mlflow.end_run(RunStatus.to_string(RunStatus.RUNNING))

        model_artifact_log = ModelArtifactRunLog(
            artifact_path=artifact_path, framework=framework
        )
        run_log = model_artifact_log.to_run_log(key=model_name, step=step)
        self.mlflow_client.insert_run_logs(run_uuid=self.run_id, run_logs=[run_log])
        logger.info("Model logged successfully")

    def get_model(
        self,
        dest_path: typing.Optional[str] = None,
        step: typing.Optional[int] = None,
        **kwargs,
    ):
        """
        Deserialize and return the logged model object for the current `run`
            and given step, returns the latest logged model(one logged at the
            largest step) if step is not passed

        Args:
            dest_path (Optional[str], optional): The path where the model is
            downloaded before deserializing.
            step (int, optional): step/iteration at which the model is being logged
                If not passed, the latest logged model (model logged with the
                highest step) is returned
            kwargs: Keyword arguments to be passed to the de-serializer.
        """
        model_name = DEFAULT_MODEL_NAME

        if step is None:
            run_log = self._get_latest_run_log(model_name)
        else:
            run_log = self._get_run_log(model_name, step)
        if run_log is None:
            raise MlFoundryException("Model is not logged")

        model_artifact_log = ModelArtifactRunLog.from_run_log(run_log)
        model_uri = self.mlflow_client.download_artifacts(
            self.run_id, path=model_artifact_log.artifact_path, dst_path=dest_path
        )
        model_framework = get_model_registry(model_artifact_log.framework)
        return model_framework.load_model(
            model_uri,
            **kwargs,
        )

    def download_model(self, dest_path: str, step: typing.Optional[int]):
        """
        Download logged model for the current `run` at a particular `step` in a
            local directory. Downloads the latest logged run (one logged at the
            largest step) if step is not passed.xs


        Args:
            dest_path (str): local directory where the model will be downloaded.
                if `dest_path` does not exist, it will be created.
                If dest_path already exist, it should be an empty directory.
            step (int, optional): step/iteration at which the model is being logged
                If not passed, the latest logged model (model logged with the
                highest step) is downloaded
        """
        model_name = DEFAULT_MODEL_NAME
        if step is None:
            run_log = self._get_latest_run_log(model_name)
        else:
            run_log = self._get_run_log(model_name, step)

        if run_log is None:
            raise MlFoundryException("Model is not logged")

        model_artifact_log = ModelArtifactRunLog.from_run_log(run_log)
        if not os.path.exists(dest_path):
            os.makedirs(dest_path)
        elif any(os.scandir(dest_path)):
            raise MlFoundryException("model download path should be empty")
        with tempfile.TemporaryDirectory() as local_dir:
            self.mlflow_client.download_artifacts(
                self.run_id, path=model_artifact_log.artifact_path, dst_path=local_dir
            )
            model_dir = os.path.join(
                local_dir, os.path.normpath(model_artifact_log.artifact_path)
            )
            copy_tree(model_dir, dest_path)



"""
# TO independently test this module, you can run the example in the path
python examples/sklearn/iris_train.py

Besides running pytest
"""
import datetime
import logging
import os
import urllib
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Sequence

import coolname
import mlflow
import pandas as pd
from mlflow.store.artifact.artifact_repository_registry import (
    _artifact_repository_registry,
)
from mlflow.store.tracking import SEARCH_MAX_RESULTS_DEFAULT
from mlflow.tracking import MlflowClient

from mlfoundry import amplitude, constants
from mlfoundry.artifact.truefoundry_artifact_repo import TruefoundryArtifactRepository
from mlfoundry.enums import ViewType
from mlfoundry.exceptions import MlflowException, MlFoundryException
from mlfoundry.internal_namespace import NAMESPACE
from mlfoundry.mlfoundry_run import MlFoundryRun
from mlfoundry.run_utils import append_path_to_rest_tracking_uri, resolve_tracking_uri
from mlfoundry.session import Session
from mlfoundry.tracking.auth_service import AuthService
from mlfoundry.tracking.truefoundry_rest_store import get_rest_store

if TYPE_CHECKING:
    from mlfoundry.inference.store import ValueType

logger = logging.getLogger(__name__)


def init_rest_tracking(tracking_uri: str, api_key: Optional[str]):
    """init_rest_tracking.

    Args:
        tracking_uri (str): tracking_uri
        api_key (Optional[str]): api_key
    """
    rest_store = get_rest_store(tracking_uri)
    auth_service = AuthService(tenant_info=rest_store.get_tenant_info())

    session = Session(auth_service=auth_service, tracking_uri=tracking_uri)
    session.init_session(api_key)

    artifact_repository = partial(TruefoundryArtifactRepository, rest_store=rest_store)
    _artifact_repository_registry.register("s3", artifact_repository)


def get_client(
    tracking_uri: Optional[str] = None,
    inference_store_uri: Optional[str] = None,
    disable_analytics: bool = False,
    api_key: Optional[str] = None,
) -> "MlFoundry":
    """Initializes and returns the mlfoundry client.

    Args:
        tracking_uri (Optional[str], optional): Custom tracking server URL.
            If not passed, by default all the run details are sent to Truefoundry server
            and can be visualized at https://app.truefoundry.com/mlfoundry.
            Tracking server URL can be also configured using the `MLF_HOST`
            environment variable. In case environment variable and argument is passed,
            the URL passed via this argument will take precedence.
        disable_analytics (bool, optional): To turn off usage analytics collection, pass `True`.
            By default, this is set to `False`.
        api_key (Optional[str], optional): API key.
            API key can be found at https://app.truefoundry.com/settings. API key can be
            also configured using the `MLF_API_KEY` environment variable. In case the
            environment variable and argument are passed, the value passed via this argument
            will take precedence.

    Returns:
        MlFoundry: Instance of `MlFoundry` class which represents a `run`.

    Examples:
    ### Get client
    Set the API key using the `MLF_API_KEY` environment variable.
    ```
    export MLF_API_KEY="MY-API_KEY"
    ```

    We can then initialize the client, the API key will be picked up from the
    environment variable.
    ```python
    import mlfoundry

    client = mlfoundry.get_client()
    ```

    ### Get client with API key passed via argument.
    ```python
    import mlfoundry

    API_KEY = "MY-API_KEY"
    client = mlfoundry.get_client(api_key=API_KEY)
    ```
    """
    # TODO (chiragjn): Will potentially need to make MlFoundry (and possibly MlFoundryRun) a Singleton instance.
    #                  Since this sets the tracking URI in global namespace, if someone were to call
    #                  get_client again with different tracking uri, the ongoing run's data will start getting
    #                  pushed to another datastore. Or we should not allow passing in tracking URI and just have
    #                  fixed online and offline clients
    amplitude.init(disable_analytics)
    tracking_uri = resolve_tracking_uri(tracking_uri)
    if tracking_uri.startswith("file:"):
        tracking_uri = os.path.join(tracking_uri, constants.MLRUNS_FOLDER_NAME)
    else:
        tracking_uri = append_path_to_rest_tracking_uri(tracking_uri)
        init_rest_tracking(tracking_uri=tracking_uri, api_key=api_key)
    amplitude.track(
        amplitude.Event.GET_CLIENT,
        # tracking whether user is using file:// or https://
        event_properties={
            "tracking_scheme": urllib.parse.urlparse(tracking_uri).scheme
        },
    )
    return MlFoundry(tracking_uri, inference_store_uri=inference_store_uri)


class MlFoundry:
    """MlFoundry."""

    def __init__(
        self,
        tracking_uri: Optional[str] = None,
        inference_store_uri: Optional[str] = None,
    ):
        """__init__.

        Args:
            tracking_uri (Optional[str], optional): tracking_uri
            inference_store_uri (Optional[str], optional): inference_store_uri
        """
        try:
            mlflow.set_tracking_uri(tracking_uri)
        except MlflowException as e:
            err_msg = (
                f"Could not initialise mlfoundry object. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

        self.mlflow_client = MlflowClient()
        if inference_store_uri is not None:
            from mlfoundry.inference.store import (
                InferenceStoreClient,
                get_inference_store,
            )

            self.inference_store_client: Optional[
                InferenceStoreClient
            ] = InferenceStoreClient(lambda: get_inference_store(inference_store_uri))
        else:
            self.inference_store_client = None

    def _get_or_create_project(self, project_name: str, owner: Optional[str]) -> str:
        """_get_or_create_experiment.

        Args:
            project_name (str): The name of the project.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                the current user will be used as owner. If the given owner
                does not have the project, it will be created under
                the current user.

        Returns:
            str: The id of the project.
        """
        experiment_name = project_name
        try:
            experiment = self.mlflow_client.get_experiment_by_name(
                experiment_name, owner_subject_id=owner
            )
            if experiment is not None:
                return experiment.experiment_id
            if not owner:
                logger.info(
                    f"project {experiment_name} does not exist. Creating {experiment_name}."
                )
                return self.mlflow_client.create_experiment(experiment_name)
            else:
                logger.warning(
                    f"project {experiment_name} under owner {owner} does not exist. "
                    "looking for project under current user."
                )
                return self._get_or_create_project(
                    project_name=project_name, owner=None
                )
        except MlflowException as e:
            err_msg = (
                f"Error happened in creating or getting project based on project name: "
                f"{experiment_name}. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

    def get_all_projects(self) -> List[str]:
        """Returns a list of project ids accessible by the current user.

        Returns:
            List[str]: A list of project ids.
        """
        amplitude.track(amplitude.Event.GET_ALL_PROJECTS)
        try:
            experiments = self.mlflow_client.list_experiments(view_type=ViewType.ALL)
        except MlflowException as e:
            err_msg = (
                f"Error happened in fetching project names. Error details: {e.message}"
            )
            raise MlFoundryException(err_msg) from e

        projects = []
        for e in experiments:
            # Experiment ID 0 represents default project which we are removing.
            if e.experiment_id != "0":
                projects.append(e.name)

        return projects

    def create_run(
        self,
        project_name: str,
        run_name: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        owner: Optional[str] = None,
        log_system_metrics: bool = True,
        **kwargs,
    ) -> MlFoundryRun:
        """Initialize a `run`.

        In a machine learning experiment `run` represents a single experiment
        conducted under a project.
        Args:
            project_name (str): The name of the project under which the run will be created.
                project_name should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
                The user must have `ADMIN` or `WRITE` access to this project.
            run_name (Optional[str], optional): The name of the run. If not passed, a randomly
                generated name is assigned to the run. Under a project, all runs should have
                a unique name. If the passed `run_name` is already used under a project, the
                `run_name` will be de-duplicated by adding a suffix.
                run name should only contain alphanumerics (a-z,A-Z,0-9) or hyphen (-).
            tags (Optional[Dict[str, Any]], optional): Optional tags to attach with
                this run. Tags are key-value pairs.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                the current user will be used as owner. If the given owner
                does not have the project, it will be created under
                the current user.
            log_system_metrics (bool, optional): Automatically collect system metrics. By default, this is
                set to `True`.
            kwargs:

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.

        Examples:
        ### Create a run under current user.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()

        tags = {"model_type": "svm"}
        run = client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel", tags=tags
        )

        run.end()
        ```

        ### Creating a run using context manager.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()
        with client.create_run(
            project_name="my-classification-project", run_name="svm-with-rbf-kernel"
        ) as run:
            # ...
            # Model training code
            ...
        # `run` will be automatically marked as `FINISHED` or `FAILED`.
        ```

        ### Create a run in a project owned by a different user.
        ```python
        import mlfoundry

        client = mlfoundry.get_client()

        tags = {"model_type": "svm"}
        run = client.create_run(
            project_name="my-classification-project",
            run_name="svm-with-rbf-kernel",
            tags=tags,
            owner="bob",
        )
        run.end()
        ```
        """
        amplitude.track(amplitude.Event.CREATE_RUN)
        if not run_name:
            run_name = coolname.generate_slug(2)
            logger.info(
                f"No run_name given. Using a randomly generated name {run_name}."
                " You can pass your own using the `run_name` argument"
            )
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name!r}"
            )

        experiment_id = self._get_or_create_project(project_name, owner=owner)

        if tags is not None:
            NAMESPACE.validate_namespace_not_used(tags.keys())
        else:
            tags = {}

        run = self.mlflow_client.create_run(experiment_id, name=run_name, tags=tags)
        mlf_run_id = run.info.run_id

        mlf_run = MlFoundryRun(
            experiment_id, mlf_run_id, log_system_metrics=log_system_metrics, **kwargs
        )
        # TODO(Rizwan): Revisit this once run lifecycle is formalised
        mlf_run._add_git_info()
        mlf_run._add_python_mlf_version()
        logger.info(f"Run {run.info.fqn!r} has started.")
        return mlf_run

    def get_run(self, run_id: str) -> MlFoundryRun:
        """Get an existing `run` by the `run_id`.

        Args:
            run_id (str): run_id or fqn of an existing `run`.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.
        """
        amplitude.track(amplitude.Event.GET_RUN)
        if run_id == "" or (not isinstance(run_id, str)):
            raise MlFoundryException(
                f"run_id must be string type and not empty. "
                f"Got {type(run_id)} type with value {run_id}"
            )
        if "/" in run_id:
            return self.get_run_by_fqn(run_id)

        run = self.mlflow_client.get_run(run_id)
        return MlFoundryRun(
            experiment_id=run.info.experiment_id,
            run_id=run.info.run_id,
            log_system_metrics=False,
        )

    def get_run_by_fqn(self, run_fqn: str) -> MlFoundryRun:
        """Get an existing `run` by `fqn`.

        `fqn` stands for Fully Qualified Name. A run `fqn` has the following pattern:
        owner/project_name/run_name

        If user `bob` has created a run `svm` under the project `cat-classifier`,
        the `fqn` will be `bob/cat-classifier/svm`.

        Args:
            run_fqn (str): `fqn` of an existing run.

        Returns:
            MlFoundryRun: An instance of `MlFoundryRun` class which represents a `run`.
        """
        run = self.mlflow_client.get_run_by_fqn(run_fqn)
        return MlFoundryRun(
            experiment_id=run.info.experiment_id,
            run_id=run.info.run_id,
            log_system_metrics=False,
        )

    def get_all_runs(
        self, project_name: str, owner: Optional[str] = None
    ) -> pd.DataFrame:
        """Returns all the run name and id present under a project.

        The user must have `READ` access to the project.
        Args:
            project_name (str): Name of the project.
            owner (Optional[str], optional): Owner of the project. If owner is not passed,
                                   the current user will be used as owner.
        Returns:
            pd.DataFrame: dataframe with two columns- run_id and run_name
        """
        amplitude.track(amplitude.Event.GET_ALL_RUNS)
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name}"
            )
        experiment = self.mlflow_client.get_experiment_by_name(
            project_name, owner_subject_id=owner
        )
        if experiment is None:
            return pd.DataFrame(
                columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
            )

        experiment_id = experiment.experiment_id

        try:
            all_run_infos = self.mlflow_client.list_run_infos(
                experiment_id, run_view_type=ViewType.ALL
            )
        except MlflowException as e:
            err_msg = f"Error happened in while fetching runs for project {project_name}. Error details: {e.message}"
            raise MlFoundryException(err_msg) from e

        runs = []

        for run_info in all_run_infos:
            try:
                run = self.mlflow_client.get_run(run_info.run_id)
                run_name = run.info.name or run.data.tags.get(
                    constants.RUN_NAME_COL_NAME, ""
                )
                runs.append((run_info.run_id, run_name))
            except MlflowException as e:
                logger.warning(
                    f"Could not fetch details of run with run_id {run_info.run_id}. "
                    f"Skipping this one. Error details: {e.message}. "
                )

        return pd.DataFrame(
            runs, columns=[constants.RUN_ID_COL_NAME, constants.RUN_NAME_COL_NAME]
        )

    def search_runs(
        self,
        project_name: str,
        filter_string: str = "",
        run_view_type: str = "ACTIVE_ONLY",
        order_by: Sequence[str] = ("attribute.start_time DESC",),
        owner: Optional[str] = None,
    ) -> Generator[MlFoundryRun, None, None]:
        """
        The user must have `READ` access to the project.
        Returns an Generator that returns a MLFoundryRun on each next call.
        All the runs under a project which matches the filter string and the run_view_type are returned.

        Args:
            project_name (str): Name of the project.
            filter_string (str, optional):
                Filter query string, defaults to searching all runs. Identifier required in the LHS of a search expression.
                Signifies an entity to compare against. An identifier has two parts separated by a period: the type of the entity and the name of the entity.
                The type of the entity is metrics, params, attributes, or tags. The entity name can contain alphanumeric characters and special characters.
                You can search using two run attributes : status and artifact_uri. Both attributes have string values.
                When a metric, parameter, or tag name contains a special character like hyphen, space, period, and so on,
                enclose the entity name in double quotes or backticks, params."model-type" or params.`model-type`

            run_view_type (str, optional): one of the following values "ACTIVE_ONLY", "DELETED_ONLY", or "ALL" runs.
            order_by (List[str], optional):
                List of columns to order by (e.g., "metrics.rmse"). Currently supported values
                are metric.key, parameter.key, tag.key, attribute.key. The ``order_by`` column
                can contain an optional ``DESC`` or ``ASC`` value. The default is ``ASC``.
                The default ordering is to sort by ``start_time DESC``.
            owner (Optional[str], optional):
                Owner of the project. If owner is not passed, the current user will be used as owner.

        Examples:
            ```python
            import mlfoundry as mlf

            client = mlf.get_client()
            with client.create_run(project_name="my-project", run_name="run-1") as run1:
                run1.log_metrics(metric_dict={"accuracy": 0.74, "loss": 0.6})
                run1.log_params({"model": "LogisticRegression", "lambda": "0.001"})

            with client.create_run(project_name="my-project", run_name="run-2") as run2:
                run2.log_metrics(metric_dict={"accuracy": 0.8, "loss": 0.4})
                run2.log_params({"model": "SVM"})

            # Search for the subset of runs with logged accuracy metric greater than 0.75
            filter_string = "metrics.accuracy > 0.75"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7
            filter_string = "metrics.accuracy > 0.7"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and model="LogisticRegression"
            filter_string = "metrics.accuracy > 0.7 and params.model = 'LogisticRegression'"
            runs = client.search_runs(project_name="my-project", filter_string=filter_string)

            # Search for the subset of runs with logged accuracy metric greater than 0.7 and order by accuracy in Descending  order
            filter_string = "metrics.accuracy > 0.7"
            order_by = ["metric.accuracy DESC"]
            runs = client.search_runs(
                project_name="my-project", filter_string=filter_string, order_by=order_by
            )

            ```

        Returns:
            Genarator[MlFoundryRun, None, None]: MLFoundryRuns matching the search query.
        """
        if project_name == "" or (not isinstance(project_name, str)):
            raise MlFoundryException(
                f"project_name must be string type and not empty. "
                f"Got {type(project_name)} type with value {project_name!r}"
            )
        try:
            run_view_type = ViewType.from_string(run_view_type.lower())
        except Exception as e:
            raise MlFoundryException(e) from e

        try:
            experiment = self.mlflow_client.get_experiment_by_name(
                project_name, owner_subject_id=owner
            )
        except MlflowException as e:
            raise MlFoundryException(e) from e  # user doesnot have READ permission

        if experiment is None:
            owner = f"owner - {owner}" if owner is not None else "current user"
            logger.warning(
                f"Project with name {project_name} does not exist under the {owner}."
            )
            return

        experiment_id = experiment.experiment_id

        page_token = None
        done = False
        while not done:
            all_runs = self.mlflow_client.search_runs(
                experiment_ids=[experiment_id],
                filter_string=filter_string,
                run_view_type=run_view_type,
                max_results=SEARCH_MAX_RESULTS_DEFAULT,
                order_by=order_by,
                page_token=page_token,
            )
            page_token = all_runs.token
            for run in all_runs:
                yield MlFoundryRun(
                    run.info.experiment_id, run.info.run_id, log_system_metrics=False
                )
            done = page_token is None

    @staticmethod
    def get_tracking_uri():
        """get_tracking_uri."""
        return mlflow.tracking.get_tracking_uri()

    def log_prediction(
        self,
        model_name: str,
        model_version: str,
        inference_id: str,
        features: "ValueType",
        predictions: "ValueType",
        raw_data: Optional["ValueType"] = None,
        actuals: Optional["ValueType"] = None,
        occurred_at: Optional[int] = None,
    ):
        """log_prediction.

        Args:
            model_name (str): model_name
            model_version (str): model_version
            inference_id (str): inference_id
            features (ValueType): features
            predictions (ValueType): predictions
            raw_data (Optional[ValueType]): raw_data
            actuals (Optional[ValueType]): actuals
            occurred_at (Optional[int]): occurred_at
        """
        from mlfoundry.inference.store import InferencePacket

        if self.inference_store_client is None:
            raise MlFoundryException(
                "Pass inference_store_uri in get_client function to use log_prediction"
            )
        if occurred_at is None:
            occurred_at = datetime.datetime.utcnow()
        elif not isinstance(occurred_at, int):
            raise TypeError("occurred_at should be unix epoch")
        else:
            occurred_at = datetime.datetime.utcfromtimestamp(occurred_at)
        inference_packet = InferencePacket(
            model_name=model_name,
            model_version=model_version,
            features=features,
            predictions=predictions,
            inference_id=inference_id,
            raw_data=raw_data,
            actuals=actuals,
            occurred_at=occurred_at,
        )
        self.inference_store_client.log_predictions([inference_packet])

    def log_actuals(self, model_name: str, inference_id: str, actuals: "ValueType"):
        """log_actuals.

        Args:
            model_name (str): model_name
            inference_id (str): inference_id
            actuals (ValueType): actuals
        """
        from mlfoundry.inference.store import ActualPacket

        if self.inference_store_client is None:
            raise MlFoundryException(
                "Pass inference_store_uri in get_client function to use log_prediction"
            )
        actuals_packet = ActualPacket(
            model_name=model_name, inference_id=inference_id, actuals=actuals
        )
        self.inference_store_client.log_actuals([actuals_packet])



from mlfoundry.enums import ModelFramework
from mlfoundry.log_types.pydantic_base import PydanticBase


class ModelArtifactRunLog(PydanticBase):
    artifact_path: str
    framework: ModelFramework

    @staticmethod
    def get_log_type() -> str:
        return "model"



import logging
import os
import typing

from mlfoundry import env_vars
from mlfoundry.login import get_stored_api_key
from mlfoundry.tracking.auth_service import AuthService

logger = logging.getLogger(__name__)


class Session:
    def __init__(self, auth_service: AuthService, tracking_uri: str):
        self.auth_service: AuthService = auth_service
        self.tracking_uri = tracking_uri

    def init_session(
        self,
        api_key: typing.Optional[str] = None,
    ):
        final_api_key = (
            api_key
            or os.getenv(env_vars.API_KEY)
            or get_stored_api_key(self.tracking_uri)
        )

        if final_api_key is None:
            # if API key is not present,
            # then take a look if MLFLOW_TRACKING_TOKEN itself has been set.
            # this will be used in sfy.
            existing_token = os.getenv(env_vars.TRACKING_TOKEN, "")
            if existing_token:
                logger.info("API key is not present. Using existing tracking token")
                return
            logger.info(
                "Session was not set as api key was neither passed, not set via env var"
            )
            return

        token = self.auth_service.get_token(api_key=final_api_key)
        os.environ[env_vars.TRACKING_TOKEN] = token



from .abstract_inference_store import AbstractInferenceStore
from .sqlalchemy_inference_store import SqlAlchemyInferenceStore


def get_inference_store(uri: str) -> AbstractInferenceStore:
    # For now we have only one repo
    return SqlAlchemyInferenceStore(db_uri=uri)



import abc
import typing

from ..inference_packet import BasePacket


class AbstractInferenceStore(abc.ABC):
    @abc.abstractmethod
    def batch_log_inference(self, inference_packets: typing.List[BasePacket]):
        ...

    @abc.abstractmethod
    def list_inference(
        self,
        model_name: str,
        model_version: typing.Optional[str] = None,
        page: int = 0,
        page_size: typing.Optional[int] = None,
    ):
        ...
